/*!***************************************************************************
*! FILE NAME  : main.c
*! DESCRIPTION: Program to write captured video (and audio) to camera file system
*! using Matroska container.
*! Copyright (C) 2010 Elphel, Inc.
*! -----------------------------------------------------------------------------**
*!  This program is free software: you can redistribute it and/or modify
*!  it under the terms of the GNU General Public License as published by
*!  the Free Software Foundation, either version 3 of the License, or
*!  (at your option) any later version.
*!
*!  This program is distributed in the hope that it will be useful,
*!  but WITHOUT ANY WARRANTY; without even the implied warranty of
*!  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*!  GNU General Public License for more details.
*!
*!  You should have received a copy of the GNU General Public License
*!  along with this program.  If not, see <http://www.gnu.org/licenses/>.
*! -----------------------------------------------------------------------------**
*!
*!  $Log$
*/
#define __USE_POSIX199506 
#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <fcntl.h>
#include <inttypes.h>
#include <pthread.h>
#include <signal.h>
#include <string.h>
#include <errno.h>
#include <sys/mman.h>
#include <sys/stat.h>

#include <asm/elphel/c313a.h>

#include "config.h"
#include "cammkv_cmd.h"
#include "defines.h"

///////////////////////////////// Global data ///////////////////////////////////////////
int   			debug_level = 0;
FILE			*debug_file;

unsigned long 		*ccam_dma_buf;           /* mmapped array */
struct framepars_all_t  *frameParsAll;
struct framepars_t      *framePars;
unsigned long           *globalPars; 	// parameters that are not frame-related, 
					// their changes do not initiate any actions
_state_t 		sstate;
_state_t 		*state = &sstate;

pthread_mutex_t thread_lock 	= PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t mkv_lock 	= PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t stopping_lock 	= PTHREAD_MUTEX_INITIALIZER;
pthread_cond_t  cond  		= PTHREAD_COND_INITIALIZER;


/////////////////////////////////////////////////////////////////////////////////////////

void quit(int n) 
{
	if(state->running==0 && state->starting==0)
		exit(0);
	DBG(3,"Signal quitting ... \n");
	cammkv_stop();
	cammkv_free();
	state->done=1;
	pthread_mutex_unlock (&thread_lock);
}

int open_handlers(void)
{
	const char 	ExifFileName[]		= "/dev/exif_exif";
	const char 	HeadFileName[]		= "/dev/jpeghead";
	const char 	ctlFileName[]		= "/dev/frameparsall";
	const char 	circbufFileName[]	= "/dev/circbuf";

	//! open Exif header file
	state->fd_exif = open(ExifFileName, O_RDONLY);
	if (state->fd_exif<0) { 	// check control OK
		DBG(0,"Error opening %s\n", ExifFileName);
		return -1;
	}
	//! open JPEG header file
	state->fd_head = open(HeadFileName, O_RDWR);
	if (state->fd_head<0) { 	// check control OK
		DBG(0,"Error opening %s\n", HeadFileName);
		return -1;
	}
	state->head_size=lseek(state->fd_head, 0, SEEK_END);
	if (state->head_size>JPEG_HEADER_MAXSIZE) {
		DBG(0,"%s:%d: Too big JPEG header (%d > %d)",__FILE__,__LINE__,\
				state->head_size, JPEG_HEADER_MAXSIZE );
		return -2;
	}

	//! open circbuf and mmap it (once at startup)
	state->fd_circ = open(circbufFileName, O_RDWR);
	if (state->fd_circ<0) { 	// check control OK
		DBG(0,"Error opening %s\n", circbufFileName);
		return -2;
	}
	/*! find total buffer length (it is in defines, actually in c313a.h */
	state->circ_buff_size=lseek(state->fd_circ, 0, SEEK_END);
	ccam_dma_buf = (unsigned long *) mmap(0, state->circ_buff_size, PROT_READ, MAP_SHARED, state->fd_circ, 0);
	if((int)ccam_dma_buf == -1) {
		DBG(0,"Error in mmap of %s\n",circbufFileName);
		close(state->fd_circ);
		return -3;
	}

	//! Now open/mmap file to read sensor/compressor parameters 
	//  (currently - just free memory in circbuf and compressor state)

	//! open circbuf and mmap it (once at startup)
	state->fd_fparmsall = open(ctlFileName, O_RDWR);
	if (state->fd_fparmsall<0) { // check control OK
		DBG(0,"%s:%d:%s: Error opening %s\n",__FILE__,__LINE__,__FUNCTION__, ctlFileName);
		return -2;
	}

	//! now try to mmap
	frameParsAll = (struct framepars_all_t *) mmap(0, sizeof (struct framepars_all_t) , 
			PROT_READ, MAP_SHARED, state->fd_fparmsall, 0);
	if((int)frameParsAll == -1) {
		DBG(0,"%s:%d:%s: Error in mmap in %s\n",__FILE__,__LINE__,__FUNCTION__,ctlFileName);
		close(state->fd_fparmsall);
		close(state->fd_circ);
		return -3;
	} 
	framePars=frameParsAll->framePars;
	globalPars=frameParsAll->globalPars;

return 0;
}

int main_loop(FILE *cmd_file)
{
	int 			cmd;
	int			res;
	
	//! Here is a main loop. If recording is on, it will check for commands after each frame, 
	//  if it is off - poll with fixed usleep
	while (!state->done) {
		//! look at command queue first
		if((cmd=parse_cmd(cmd_file))) {
			if (cmd<0) DBG(0,"Unrecognized command\n");
		} else if (state->running) { //!no commands in queue, started
			// here real working
			usleep( COMMAND_LOOP_DELAY) ; //! make it longer but interruptible by signals?
		} else if (state->starting) { //!no commands in queue,starting (but not started yet)
			//!retry starting
			DBG(4,"Retry starting...\n");
			switch ((res=-cammkv_start())) {
				case 0:
					//!Now we are ready to initialize MKV file
					DBG(3,"Subsytems inited! Try opening MKV output file\n");
					if(state->video_format==CAMMKV_FORMAT_MKV) 
						res=dvr_mkvopen();
					else
						res=dvr_movopen();
					if(res<0)
						return -1;
					DBG(3,"MKV file opened\n");
					state->running=1;
					state->starting=0;
					DBG(3,"Unlock mutex and enable threads.\n");
					pthread_mutex_unlock (&thread_lock);
					DBG(3,"Mutex unlocked\n");

					break; //! file started OK, nothing to do
				case CAMMKV_FRAME_NOT_READY:        
					//!  just wait for the frame to appear at the current pointer
					//! we'll wait for a frame, not to waste resources. 
					//! But if the compressor is stopped this program will not respond to any commands
					//! TODO - add another wait with (short) timeout?
					//                lseek(state->fd_circ,LSEEK_CIRC_WAIT,SEEK_END);
				case  CAMMKV_FRAME_CHANGED:   //! frame parameters have changed
				case  CAMMKV_FRAME_NEXTFILE:
				case  CAMMKV_FRAME_INVALID:   //! invalid frame pointer
				case  CAMMKV_FRAME_BROKEN:    //! frame broken (buffer overrun)
					usleep( COMMAND_LOOP_DELAY) ; //! it should be not too long so empty buffer will not be overrun
					break;
				case  CAMMKV_FRAME_FILE_ERR:  //! error with file I/O
				case  CAMMKV_FRAME_OTHER:     //! other errors
					DBG(0,"%s:line %d - error=%d\n",__FILE__,__LINE__,res);
					break;
				default:
					DBG(0,"%s:line %d - should not get here (rslt=%d)\n",__FILE__,__LINE__,res);
					return (-1);
			} //switch
		} else { //! not running, not starting
			usleep( COMMAND_LOOP_DELAY) ; //! make it longer but interruptible by signals?
		}
	}
       DBG(3,"Exit from main loop\n");	
return 0;
}


static void usage(char *appl) 
{
	printf("\n Utility %s version %s. Elphel Inc, 2010 (C)",appl,VERSION);
	printf("\n\n Starting:\t%s [options] <pipe_name>\n", appl);
	printf("\noptions are:\n\t\t -S <shed type>\ttype of scheduling threads:\n");
	printf("\n\t\t\t 0\t- for SCHED_OTHER(default)");
	printf("\n\t\t\t 1\t- for SCHED_FIFO");
	printf("\n\t\t\t 2\t- for SCHED_RR");
	printf("\n\t\t -A <prio>\tpriority for audio thread.");
	printf("\n\t\t -V <prio>\tpriority for video thread.");
	printf("\n\t\t---------------------------------------------------------------------------");
	printf("\n\t\t\t 0 \t- used for all threads if SCHED_OTHER dispatcher(default)");
	printf("\n\t\t\t 1-99\t- for SCHED_FIFO or SCHED_RR dispatcher\n");
	printf("\n\t\tIf audio priority not entered: audio priority = video priority + 5\n");
	printf("\n\t\t---------------------------------------------------------------------------");
	printf("\n\t\t\tFor SCHED_FIFO or SCHED_RR you must have root privileges!");
        printf("\n\t\tOther way - use sudo or set sticky bit:\n\t\t\t chown root %s && chmod +s %s\n\t\tbefore run program\n\n",appl,appl);
return;	
}



void cammkv_init(void)
{
	bzero(state, sizeof(_state_t));
	cammkv_set_segment_duration(DEFAULT_DURATION);
	cammkv_set_segment_length(DEFAULT_LENGTH);
	cammkv_set_max_frames(DEFAULT_FRAMES);
	cammkv_set_frames_per_chunk(DEFAULT_FRAMES_PER_CHUNK);
	cammkv_set_timescale(1.0); 
	cammkv_reset(); //! sets    state->buf_overruns=-1; //!first does not count
	cammkv_set_prefix ("/var/hdd/");
	cammkv_set_debug_out("stderr");
	cammkv_set_frames_skip(0); //! don't skip
	cammkv_set_audio_state("on"); // enable audio
	set_audio_pars();
	cammkv_set_exif(0);
	cammkv_set_format(CAMMKV_FORMAT_MKV);
	///kml stuff
	cammkv_kml_set_enable(0);
	state->kml_file=NULL;
	state->kml_path[0]='\0';
        cammkv_kml_set_horHalfFov(20.0);
	cammkv_kml_set_vertHalfFov(15.0);
	cammkv_kml_set_near(40.0); // distance to PhotoOverlay
	cammkv_kml_set_period(2);
	cammkv_kml_set_height(10.0);
	cammkv_kml_set_height_mode(0);
}

int cammkv_start(void)
{
	int	res;

	res = init_video();
	if(res != 0) return res;
	if(state->audio_enabled) init_audio();
	DBG(3,"cammkv inited!\n");
return 0;
}

void cammkv_stop(void)
{
	DBG(3,"cammkv_stop\n");
	if(!state->running)
		return;
	pthread_mutex_lock (&thread_lock);
	stopping_audio();
	deinit_audio();
	if( state->running) {
		stopping_video();
		state->running=0;
		state->starting=0;
		deinit_video();
	}
	cammkv_end_kml();
	state->cirbuf_rp=-1;
}




int main(int argc, char *argv[]) 
{
	int			c,flag = 0;
	int			res,i;
	int			defpri_v   = 0;   // video thread priority
	int			defpri_a   = 0;   // audio thread priority
	pthread_attr_t 		atr;
	pthread_t		th_a, th_v;
	struct sched_param 	sp; 
	int 			po = SCHED_OTHER; 
	FILE 			*cmd_file;
	char			pipe_name[128];


	cammkv_init(); // init default values
	signal(SIGINT, quit);
	// parse command line parameters
	while(1) {
		c = getopt(argc, argv, "S:V:A:h?");
		if (c == -1)
			break;
		switch (c) {
			case 'S':
				po = atoi(optarg);
				if(!(po>=0 && po<3)) {
					flag++;
					DBG(0,"error: Incorrect type of scheduling [%d]! Value must be from 0 to 2.\n",po);
				}
				break;
			case 'V':
				defpri_v = atoi(optarg);
				if(!(defpri_v >= 0 && defpri_v < 100)) {
					flag++;	
					DBG(0,"error: Incorrect priority settings for video thread! Value must be from 0 to 99\n");
				}
				break;
			case 'A':
				defpri_a = atoi(optarg);
				if(!(defpri_a >= 0 && defpri_a < 100)) {
					flag++;	
					DBG(0,"error: Incorrect priority settings for dumper thread! Value must be from 0 to 99\n");
				}
				break;
			case 'h':
			case '?':
			default:
				usage((char *)argv[0]);
				exit(1);

			//	DBG(0,"error: ?? getopt returned character code 0%o ??\n", c);
		}
	} //while
	if(optind>=argc) {
		printf("Expected name pipe after options.\n");
		exit(1);
	}
	strcpy(pipe_name,argv[optind]);
	pthread_mutex_lock (&thread_lock); // block threads until not init video and not open MKV file
	
	// starting video and audio consumer threads and block it
	if(po == SCHED_OTHER) {
		defpri_v=defpri_a=0;
	}
	else {
		// SCHED_FIFO and SCHED_RR - check priority threads
		if(defpri_v==0)
			defpri_v=10;
		else if(defpri_v >100)
			defpri_v=99;
		if(defpri_a==0)
			defpri_a=defpri_v+5;
		if(defpri_a >= 100)
			defpri_a=99;
	}

	DBG(5,"Set default priority for threads type %d\n",po);

	DBG(5,"Trying init video thread \n");
	pthread_attr_init( &atr ); 
	pthread_attr_setinheritsched( &atr, PTHREAD_CREATE_DETACHED );    
	pthread_attr_setinheritsched( &atr, PTHREAD_EXPLICIT_SCHED );    
	pthread_attr_setschedpolicy( &atr, po );
	sp.sched_priority=defpri_v;
	//   sp.__sched_priority = defpri_v; 
	res =  pthread_attr_setschedparam( &atr, &sp ); 
	if( res != 0 ) { 
		DBG(0,"Set priority for video thread error: %s\n ",strerror( res )); 
		exit(1);
	}
	DBG(5,"Trying  create video thread \n");
	if( pthread_create(&th_v, &atr, thread_v, (void *)state)) {
		DBG(0,"Unable to create video thread! Exiting ...\n");
		exit(1);
	}

	DBG(5,"Trying init audio thread \n");
	pthread_attr_init( &atr ); 
	pthread_attr_setinheritsched( &atr, PTHREAD_CREATE_DETACHED );    
	pthread_attr_setinheritsched( &atr, PTHREAD_EXPLICIT_SCHED );    
	pthread_attr_setschedpolicy( &atr, po ); 
	sp.sched_priority = defpri_a; 
	//   sp.__sched_priority = defpri_dmp; 
	res =  pthread_attr_setschedparam( &atr, &sp ); 
	if( res != 0 ) { 
		DBG(0,"Set priority for audio thread error: %s\n ",strerror( res )); 
		exit(1);
	}
	DBG(5,"Trying  create audio thread \n");
	if(pthread_create(&th_a, &atr, thread_a, (void *)state)) {
		DBG(0,"Unable to create audio thread! Exiting ...\n");
		exit(1);
	}
	
	if(open_handlers()) exit(1);

	//!create a named pipe
	//!always delete the pipe if it existed, start a fresh one
	i=unlink (pipe_name);
	if (i) {
		DBG(1,"Unlink %s returned %d, errno=%d \n", pipe_name, i, errno);
	}
	i=mkfifo(pipe_name, 0777); //EEXIST
	//! now should not exist
	if (i) {
		if (errno==EEXIST) {
			DBG(1,"Named pipe %s already exists, will use it.\n", pipe_name);
		} else {
			DBG(0,"Can not create a named pipe %s, errno=%d \n", pipe_name, errno);
			return -4;
		}
	}

	//!now open the pipe - will block until something will be written (or just open for writing
	//!Reads themselves will not block
	if (!((cmd_file=fopen(pipe_name,"r")))) {
		DBG(0,"Can not open command file %s\n",argv[1]);
		return -5;
	}
	DBG(1,"Pipe %s opened for reading\n",pipe_name);

	DBG(3,"Go to main loop.\n");
	main_loop(cmd_file);

	/* never returns while not interrupted*/
	pthread_join( th_v, NULL ); 
	pthread_join( th_a, NULL ); 
	DBG(3,"Threads stopped. \n");
	fclose(cmd_file);
	cammkv_free();

return 0;
}

